#include <string>
#include <iostream>
#include <math.h>
#include "TSeries.hh"
#include "FSeries.hh" 
#include "DatEnv.hh" 
#include "Trend.hh"  // Include Monitor Data API 

const char* const kTrendfile = "RmsMonitor";
const float kLowF = 10;
const float kHighF = 30;

class RmsMonitor : public DatEnv { 
   // inherit from DatEnv and MonServer 
   public: 
      RmsMonitor (int argc, const char* argv[]); 
      virtual ~RmsMonitor() {} 
      virtual void ProcessData();
      TSeries*     fData;      // Input data 
      Trend        fHistory;   // Trend class
      std::string  fTrendName; // Name of trend channel 
}; 

EXECDAT (RmsMonitor);   // DMT Framework 

// Constructs the monitor, initializes the monitor server and subscribes
// the data objects 
RmsMonitor::RmsMonitor (int argc, const char* argv[]) 
: DatEnv (argc, argv), fData (0) 
{
   if ((argc <= 1) || (argv[1] == "-h") || (argv[argc-1][0] == '-')) {
      cout << "usage: trend_example 'channel name'" << endl;
      finish();
   }
   else {
      getDacc().addChannel (argv[argc-1], 1, &fData); 
      // setup trend channel
      fTrendName = argv[argc-1];          // Name of the trend channel
      fTrendName += "_RMS";
      fHistory.setName (kTrendfile);      // Set filename of trend
      fHistory.setType (Trend::kSecond);  // Set trend type (sec or min)
      fHistory.setFrameLen (60);          // Set frame length
      fHistory.addChannel (fTrendName.c_str()); // Add trend channel
   }
} 

// Get some data and load the history buffer 
void RmsMonitor::ProcessData() 
{ 
   // Calculate the rms value 
   if (fData) {
      FSeries* fft = new FSeries (*fData); 
      float rms = sqrt (fft->Power (kLowF, kHighF)); 
      delete fft;
      fHistory.trendData (fTrendName.c_str(), fData->getStartTime(), rms);
   }
}
