#ifndef __EXTENSIONS__
#define __EXTENSIONS__
#endif
#include <time.h>
#include <pthread.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <strings.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <vector>
#include <deque>
#include <list>
#include <iostream>
#include <fstream>
#include <sstream>
#include "gmutex.hh"
#include "goption.hh"
#include "./memstream.hh"   // NEEDS TO BE FIXED
#include "sockutil.h"
#include "Interval.hh"
#include "lsmp_prod.hh"
#include "ccrc.hh"
#include "TimingCollector.hh"

using namespace std;

const char* const help_text = 
"Usage: TimingCollector [-f 'conf' -m 'smem' -s 'size' -n 'num' -l 'log']\n"
"       Collects timing information and writes them into shared memory.\n"
"       -f 'conf' : configuration file name\n"
"       -a 'ip' : address of RS422 converter\n"
"       -p 'port' : port number of RS422 converter (default is 1234)\n"
"       -m 'smem' : name of shared memory (default is TimingSystem)\n"
"       -s 'size' : size of partition in shared memory (default is 1 MB)\n"
"       -n 'num' : number of partitions (default is 8)\n"
"       -l 'file' : name of log file\n"
"       -1: run once and dump data to standard output\n"
"       -d: print debug messages"
;

//______________________________________________________________________________
typedef vector<port_collector> port_vector;
typedef list<mfo_tree> data_list;

uint32_t Reflect(uint32_t ulReflect, char cChar)
{
      uint32_t ulValue = 0;

      // Swap bit 0 for bit 7 bit 1 For bit 6, etc....
      for(int iPos = 1; iPos < (cChar + 1); iPos++)
      {
         if(ulReflect & 1) ulValue |= 1 << (cChar - iPos);
         ulReflect >>= 1;
      }

      return ulValue;
}

//______________________________________________________________________________

bool		 Debug = false;
bool		 UseLogfile = false;
ofstream 	 Logfile;
deque<string> 	 Logs;
thread::mutex 	 Logmux;
pthread_t	 LogTID;

string timestamp (const Time& t = Now())
{
   char buf[256];
   buf[0] = 0;
   TimeStr (t, buf, "%02d %03M %Y, %02H:%02N:%02S : ");
   return string (buf);
}

bool writelog (const string& s)
{
   thread::semlock lockit (Logmux);
   Logs.push_back (timestamp() + s);
   return true;
}

extern "C" {
   void* logging (void*)
   {
      while (1) {
         sleep (2);
         {
            thread::semlock lockit (Logmux);
	    while (!Logs.empty()) {
	       if (UseLogfile) {
	          Logfile << Logs.front();
	       }
	       Logs.pop_front();
	    }
	 }
      }
      return 0;
   }
}

//______________________________________________________________________________
extern "C" {
   void* read_port_global (void* port)
   {
      return ((port_collector*)port)->read_port();
   }
}

//______________________________________________________________________________
#define byteswap2(a,b) ((char *)&(a))[0] = ((char *)&(b))[1]; ((char *)&(a))[1] = ((char *)&(b))[0];
#define byteswap4(a,b) ((char *)&(a))[0] = ((char *)&(b))[3]; ((char *)&(a))[1] = ((char *)&(b))[2];((char *)&(a))[2] = ((char *)&(b))[1];((char *)&(a))[3] = ((char *)&(b))[0];
#define byteswap8(a,b) ((char *)&(a))[0] = ((char *)&(b))[7]; ((char *)&(a))[1] = ((char *)&(b))[6];((char *)&(a))[2] = ((char *)&(b))[5];((char *)&(a))[3] = ((char *)&(b))[4];((char *)&(a))[4] = ((char *)&(b))[3]; ((char *)&(a))[5] = ((char *)&(b))[2]; ((char *)&(a))[6] = ((char *)&(b))[1]; ((char *)&(a))[7] = ((char *)&(b))[0];

inline void swap (uint32_t* w)
{
   uint32_t tmp;
   byteswap4(tmp, *w);
   *w = tmp;
}
inline bool littleendian ()
{
   uint32_t test = 0; *(uint8_t*) &test = 1;
   return (test == 1);
}


//______________________________________________________________________________
// Boyer-Moore-Horspool text search for magic string
int gMagicStringLen = -1;
int gBadCharSkip[256];
char gMagicString[256];

const char* boyermoore (const char* haystack, int hlen)
{
    if (gMagicStringLen < 0) {
       strcpy (gMagicString, kMagicString);
       gMagicStringLen = strlen (kMagicString);
       for (int i = 0; i < gMagicStringLen; ++i) 
          if (gMagicString[i] == 'Z') gMagicString[i] = 0;
       /* Initialize the table to default value */
       for(int a=0; a<256; ++a) {
          gBadCharSkip[a] = gMagicStringLen;
       }
       /* Then populate it with the analysis of the needle */
       /* But ignoring the last letter */
       int last = gMagicStringLen - 1;
       for (int a = 0; a < last; ++a) {
          gBadCharSkip[(unsigned char)gMagicString[a]] = last - a;
//	  cout << a << ": " << (int) gMagicString[a]  << endl;
       }
    }
 
    if(gMagicStringLen > hlen || gMagicStringLen <= 0 || !haystack) 
        return NULL;
  
    /* Search the haystack, while the needle can still be within it. */
    int nlen = gMagicStringLen;
    int scan;
    int last = gMagicStringLen - 1;
    while (hlen >= nlen)
    {
        /* scan from the end of the needle */
        for (scan = last; haystack[scan] == gMagicString[scan]; --scan)
            if (scan == 0) /* If the first byte matches, we've found it. */
                return haystack;
 
        /* otherwise, we need to skip some bytes and start again. 
           Note that here we are getting the skip value based on the last byte
           of needle, no matter where we didn't match. So if needle is: "abcd"
           then we are skipping based on 'd' and that value will be 4, and
           for "abcdd" we again skip on 'd' but the value will be only 1.
           The alternative of pretending that the mismatched character was 
           the last character is slower in the normal case (Eg. finding 
           "abcd" in "...azcd..." gives 4 by using 'd' but only 
           4-2==2 using 'z'. */
        hlen     -= gBadCharSkip[(unsigned char)haystack[last]];
        haystack += gBadCharSkip[(unsigned char)haystack[last]];
    }

    return NULL;
}

//______________________________________________________________________________
void* port_collector::read_port ()
{
   while (1) {
      fOpened = false;
      // create the socket
      fSocket = ::socket (PF_INET, SOCK_STREAM, 0);
      if (fSocket == -1) {
         ostringstream os;
         os << "Unable to create socket for " << fPortname << endl;
	 cerr << os.str();
	 writelog (os.str());
         fError = -1;
         return 0;
      }
      // bind socket
      struct sockaddr_in socknam;
      socknam.sin_family      = AF_INET;
      socknam.sin_port        = 0;
      socknam.sin_addr.s_addr = 0;
      int len                 = sizeof(socknam);
      if (::bind (fSocket, (struct sockaddr *)&socknam, len) < 0) {
         ostringstream os;
         os << "Unable to bind socket for " << fPortname << endl;
	 cerr << os.str();
	 writelog (os.str());
         fError = -2;
         return 0;
      }
      // connect to box
      socknam.sin_family      = AF_INET;
      socknam.sin_port        = htons (fPort);
      if (nslookup (fPortname.c_str(), &socknam.sin_addr) < 0) {
         ostringstream os;
         os << "Unable to connect socket for " << fPortname << endl;
	 cerr << os.str();
	 writelog (os.str());
         fError = -3;
         return 0;
      }
      wait_time timeout = 10.0;	/* timeout */
      //timeval timeout = {10, 0};
      if (connectWithTimeout (fSocket, (struct sockaddr *)&socknam, 
                              sizeof (socknam), timeout) < 0) {
         ostringstream os;
	 os << "Unable to connect to " << fPortname << endl;
	 cerr << os.str();
	 writelog (os.str());
         // fError = -4;
         // return 0;
      }
      else {
         fOpened = true;
         ostringstream os;
	 os << "Connected to " << fPortname << " (" << fPort << ")" << endl;
	 writelog (os.str());
      }
      if (Debug) {
         cout << "Port opened at " << fPortname << " (" << fPort << ")" << endl;
      }

      // wait for data
      char buffer[64*1024];
      char* p = buffer;
      int nRead = 0;
      while (fOpened) {
         // get some data
         //if (Debug) cout << "Wait for chars... " << endl;
         int nB = ::recv (fSocket, p, sizeof (buffer)-nRead, 0);
         if (Debug) cout << "Number of chars received " << nB << " (total " << 
	     nRead + (nB > 0 ? nB : 0) << ")" << endl;
         if (nB == -1) {
            ::close (fSocket);
            fSocket = -1;
            fOpened = false;
            ostringstream os;
            os << "Lost connection on address " << 
               fPortname << " / port " << fPort << endl;
            writelog (os.str());
            break;
         }

	 // Lantronix UDS2100: adds 0x00 to 0x0D and 0xFF to 0xFF. Argh!
// 	 int j = (nRead == 0) ? 1 : 0;
// 	 for (int k = j; k < nB; ++k) {
// 	    if (((p[k-1] == (char)0x0D) && (p[k] == (char)0x00)) ||
// 	        ((p[k-1] == (char)0xFF) && (p[k] == (char)0xFF))) {
//                p[k] = (char)0x00;
// 	    }
// 	    else {
// 	       p[j++] = p[k];
// 	    }
// 	 }
// 	 if (Debug && nB != j) cout << "Lantronix fix " << nB - j << endl;
// 	 nB = j;
	 
         // Set new buffer size
         nRead += nB;
         p += nB;
	 
         // check if we have a magic string as a header
// 	 for (int i = 0; i < 36; ++i) 
// 	    printf ("%3i  %3i  0x%02X  %c\n", i, (unsigned char)buffer[i], 
// 	            (unsigned char)buffer[i], 
// 	            isprint (buffer[i]) ? buffer[i] : ' ');
	 const char* magic = boyermoore (buffer, nRead);
	 if (!magic) { // garbage!
            if (Debug && nRead > 32) cout << "No magic header in " << nRead << endl;
	    if (nRead > gMagicStringLen) {
	       int diff = nRead - gMagicStringLen;
   	       memmove (buffer, buffer + diff, nRead - diff);
	       nRead -= diff;
	       p -= diff;
               if (Debug) cout << "Skip " << diff << " chars" << endl;
	    }
	    continue;
	 }
	 // skip everything before magic string
	 if (magic != buffer) {
	    int diff = magic - buffer;
	    memmove (buffer, magic, nRead - diff);
	    nRead -= diff;
	    p -= diff;
            if (Debug) cout << "Skip leading " << diff << " chars" << endl;
	 }
//	 for (int i = 0; i < 36; ++i) 
//	    printf ("%3i  %3i  0x%02X  %c\n", i, (unsigned char)buffer[i], 
//                  (unsigned char)buffer[i], 
//	            isprint (buffer[i]) ? buffer[i] : ' ');
         // check if we have enough data
	 if (nRead < kDataLenV1 + gMagicStringLen) {
	    continue; // not enough yet
	 }
         if (Debug) cout << "Received full buffer " << nRead <<
	   "( >=" << kDataLenV1 + gMagicStringLen << ")" << endl;
	 
         // check CRC
         CCRC32 crc  (0xffffffff, true,  false);
	 CCRC32::crc_type crcval = 
	    crc.FullCRC  (buffer, kDataLenV1 + gMagicStringLen);
	 if (Debug) printf ("CRC value: 0x%08lX\n", crcval);
// 	 if (Debug) {
//   	    for (int i = 0; i < kDataLenV1 + gMagicStringLen ; ++i) {
// 	       printf ("%02X", (unsigned char)buffer[i]);
// 	    }
// 	    printf ("\n");
// 	 }
	 
// 	 for (int i = 0; i < kDataLenV1 + gMagicStringLen; ++i) {
// 	 crcval  = crc.FullCRC  (buffer, i+1);
// 	 crcval2 = crc2.FullCRC (buffer, i+1);
// 	 //if (Debug) printf ("CRC value: 0x%08lX or 0x%08lX\n", crcval, crcval2);
// 	 if (Debug) printf ("%04i: CRC value: 0x%08X or 0x%08X\n", i+1, Reflect(crcval,32),
//  	                    Reflect(crcval2,32));
// 	 }
	 // if CRC error skip to next magic
	 if (crcval != CRC32_MATCH) {
            if (Debug) cout << "CRC error" << endl;
	    const char* m = boyermoore (buffer+1, nRead-1);
	    if (m == 0) { 
	       int diff = nRead - gMagicStringLen;
   	       memmove (buffer, buffer + diff, nRead - diff);
	       nRead -= diff;
	       p -= diff;
               if (Debug) cout << "Skip " << diff << " chars" << endl;
	    }
	    else {
	       int diff = m - buffer;
   	       memmove (buffer, m, nRead - diff);
	       nRead -= diff;
	       p -= diff;
               if (Debug) cout << "Skip " << diff << endl;
            }
            ostringstream os;
	    os << "CRC Error on " << fPortname << " (" << fPort << ")" << endl;
	    writelog (os.str());
	    continue;
	 }
	 // skip header
         int diff = gMagicStringLen;
         memmove (buffer, buffer + diff, nRead - diff);
	 nRead -= diff; 
	 p -= diff;

         if (Debug) {
            CCRC32 crc  (0xffffffff, false, false);
	    printf ("slave CRCs: ");
            for (int i = 0; i < 16; ++i) {
	       if (i % 4 == 0 && i != 0) printf ("\n            ");
	       CCRC32::crc_type cval = crc.FullCRC (buffer + i*128 + 256, 128);
	       printf ("0x%08lX ", cval);
	    }
	    printf ("\n");
	 }

	 // swap byte order 
	 if (littleendian()) {
            if (Debug) cout << "Swap data" << endl;
	    uint32_t* l = (uint32_t*)buffer;
	    for  (int i = 0; i < kDataLenV1 / 4; ++i) {
	       swap (l + i);
	       if (Debug && i >= 0) printf ("%3i: 0x%08X\n", i, l[i]); 
	    }
	 }
	 
         // decode data from binary into class structure
	 const mfo_t* mfo = (mfo_t*)buffer;
//	 if (Debug) for (int i = 0; i < 16; ++i) 
//	    printf ("CRC slave (%2i): 0x%08X\n", i, mfo->CRCSlave(i));
	 
         // save data
         if (Debug) cout << "Save data" << endl;
         {
            thread::semlock lockit (fMux);
            fTime = Time (0, 0);
	    fData = *mfo;
         }

	 // discard original data
         diff = kDataLenV1;
         memmove (buffer, buffer + diff, nRead - diff);
	 nRead -= diff;
	 p -= diff;

      }
            
      // lost connection: we wait 20 sec and try again
     ::sleep (20);
   }
}

//______________________________________________________________________________
Time OvfTime;
int OvfNum = 0;

//______________________________________________________________________________
int main (int argc, char* argv[])
{

   port_vector 		Ports;  // port list
   data_list    	Data;   // data
   bool 		RunOnce;
   string		ConfFile; // congiuration file name
   bool 		UseConfFile; // is conf file speced?
   string		LogfileName;
   string 		Address; // address of RS422 box
   bool 		UseAddress; // is addressed defined?
   int			PortNum = kPORTDEFNUM; // port number of RS422 box
   string		ShMemName (kSMDEFNAME); // shared memory partition name
   int 			ShMemSize = kSMDEFSIZE;	// partition size
   int 			ShMemNum = kSMDEFNUM; // number of partitions
   char*		ShMemBuf = 0;
   Time			Last;

   // parse command line parameters
   gdsbase::option_string opt (argc, argv, "f:a:p:m:s:n:l:d1h");
   Debug = opt.opt ('d');
   RunOnce = opt.opt ('1');
   UseConfFile = opt.getOpt ('f', ConfFile);
   UseAddress = opt.getOpt ('a', Address);
   UseLogfile = opt.getOpt ('l', LogfileName);
   if (!opt.getOpt ('p', PortNum)) {
      PortNum = kPORTDEFNUM;
   }
   if (!opt.getOpt ('m', ShMemName)) {
      ShMemName = kSMDEFNAME;
   }
   if (!opt.getOpt ('s', ShMemSize)) {
      ShMemSize = kSMDEFSIZE;
   }
   if (!opt.getOpt ('n', ShMemNum)) {
      ShMemNum = kSMDEFNUM;
   }
   if (opt.help()) {
      cout << help_text << endl;
      return 0;
   }
   // cout << ConfFile << endl << UseConfFile << endl << 
   // Address << endl << UseAddress << endl << PortNum << endl <<
   // ShMemName << endl << ShMemSize << endl << ShMemNum << endl;
   
   // create shared memory region
   LSMP_PROD ShMem (ShMemName.c_str(), ShMemNum, ShMemSize);
   if (!ShMem.valid()) {
      cerr << "Unable to create shared memory partition for " << ShMemName << 
              " (Size: " << ShMemSize << ", Num: " << ShMemNum << ")" << endl;
      return 1;
   }

   // read configuration file
   if (UseConfFile) {
      ifstream inp (ConfFile.c_str());
      while (inp) {
	 char line[1024];
	 inp.getline (line, sizeof (line));
	 char* p = line;
         while (*p && strchr (" \t", *p)) ++p;
	 if (*p == '#' || !*p) {
	    continue;
	 }
	 char* comma1 = strchr (line, ',');
	 char* comma2 = 0;
	 if (comma1) {
            *comma1++ = 0;
	    while (*comma1 && strchr (" \t", *comma1)) ++comma1;
	    comma2 = strchr (comma1, ',');
            if (comma2) {
               *comma2++ = 0;
   	       while (*comma2 && strchr (" \t", *comma2)) ++comma2;
            }
	 }
         int portnum = -1;
	 if (comma1 && strchr ("0123456789", *comma1)) {
	    portnum = atoi (comma1);
	 }
	 else if (!comma2) {
	    comma2 = comma1;
	 }
	 port_vector::value_type val (line, portnum, comma2);
	 Ports.push_back (val);
      }
   }
   else if (UseAddress) {
      port_vector::value_type val (Address.c_str(), PortNum, "MFO");
      Ports.push_back (val);
   }
   if (Ports.empty()) {
      cout << help_text << endl;
      return 1;
   }
   
   // open logfile
   if (UseLogfile) {
      Logfile.open (LogfileName.c_str());
      if (!Logfile) {
         cerr << "Unable to open log file. Logging disabled." << endl;
	 UseLogfile = false;
      }
   }
   // start thread for logging
   if (UseLogfile) {
      pthread_attr_t	tattr;
      int			status;
      if (pthread_attr_init (&tattr) == 0) {
         pthread_attr_setdetachstate (&tattr, PTHREAD_CREATE_DETACHED);
         pthread_attr_setscope (&tattr, PTHREAD_SCOPE_PROCESS);
         status = pthread_create (&LogTID, &tattr, &logging, 0);
         pthread_attr_destroy (&tattr);
      }
      else {
         cerr << "Unable to start logging process" << endl;
         exit (1);
      }
   }   

   // start a thread for each port
   for (port_vector::iterator i = Ports.begin(); i != Ports.end(); ++i)
   {
      pthread_attr_t	tattr;
      int			status;
      if (pthread_attr_init (&tattr) == 0) {
         pthread_attr_setdetachstate (&tattr, PTHREAD_CREATE_DETACHED);
         pthread_attr_setscope (&tattr, PTHREAD_SCOPE_PROCESS);
         status = pthread_create (&i->fTID, &tattr, &read_port_global, &*i);
         pthread_attr_destroy (&tattr);
      }
      else {
         cerr << "Unable to start port monitor for " << i->fPortname << endl;
         exit (1);
      }
   }
      
   // store data into shared memory
   Last = Time (0);
   while (1) {
      // wait a little while (20ms)
      timespec tick = {0, 20000000};
      nanosleep (&tick, 0);
      // check if we have more recent data
      Time recent (Last);
      for (port_vector::iterator i = Ports.begin(); i != Ports.end(); ++i) {
         i->fMux.lock();
	 Time t (i->fData.mGPS);
	 i->fMux.unlock();
	 if (t > recent) {
	    recent = t;
	    char buf[256];
	    if (Debug) cout << "Found new time at " << 
	                  TimeStr (t, buf, "%02d %03M %Y, %02H:%02N:%02S") << endl;
	 }
      }
      // if not wait go back to beginning of loop
      if (recent <= Last) {
         continue;
      }
      // wait some to make sure all data has arrived
      nanosleep (&tick, 0);
      if (Debug) cout << "New data received" << endl;
      
      // copy data
      Data.clear();
      for (port_vector::iterator i = Ports.begin(); i != Ports.end(); ++i) {
         bool found = false;
         i->fMux.lock();
	 if (Time(i->fData.mGPS) == recent) {
	    Data.push_back (mfo_tree(i->fData, i->fName));
	    found = true;
	 }
	 i->fMux.unlock();
	 if (Debug && found) cout << "Get data for " << i->fName << endl;
      }
      // build tree structure
      for (data_list::iterator i = Data.begin(); i != Data.end(); ++i) {
         bool found;
         do {
	    found = false;
	    for (data_list::iterator j = Data.begin(); j != Data.end();) { 
	       if (j != i) {
                  if (i->set_fanout (*j)) {
		     if (Debug) cout << "Found a slave fanout " << j->mName << 
		          " in " << i->mName << endl;
		     j = Data.erase (j);
		     found = true;
		  }
		  else {
		     ++j;
		  }
	       } 
	       else {
	          ++j;
	       }
	    }
	 } while (found);
      }
      
      // run once?
      if (RunOnce) {
         cout << xsilMfoTreeList (Data, recent) << endl;
         exit (0);
      }

      // allocate shared memory buffer
      ShMemBuf = ShMem.get_buffer();
      if (!ShMemBuf) {
         cerr << "Unable to allocate shared memory buffer" << endl;
	 Last = recent;
         continue;
      }
      // write data to the shared memory buffer
      int n = 0; // size of buffer
      omemstream om (ShMemBuf, ShMemSize);
      om << xsilMfoTreeList (Data, recent) << endl;
      //if (Debug) cout << xsilMfoTreeList (Data, recent) << endl;
      
      n = om.rdbuf()->pubseekoff (0, ios::cur, ios_base::out);
      if (n >= ShMemSize) {
         ++OvfNum;
         if (OvfTime + Interval(10 * 60) < Now())  {
   	    OvfTime = Now();
            ostringstream os;
	    os << "Shared memory buffer overflow (" << OvfNum << ")" << endl;
	    writelog (os.str());
	 }
      }
      
      // release buffer
      ShMem.release (n);
      // go to next data buffer
      Last = recent;
   }
   
   return 0;
}
