#ifndef FILELIST_HH
#define FILELIST_HH

#include <list>
#include <string>
#include <iosfwd>

class FileList {
public:
    typedef std::list< std::string > file_list;
    typedef file_list::iterator file_iter;
    typedef file_list::const_iterator const_file_iter;

public:
    FileList(void);
    ~FileList(void);
    void addFile(const char* name);
    void addList(const char* listnm);
    bool empty(void) const;
    const char* getFront(void) const;
    unsigned long size(void) const;
    std::ostream& print(std::ostream& out) const;
    void pop_front(void);

private:
    /**  mFile contains the input data file name(s)
      *  @memo Input data file name.
      */
    file_list mFile;
};

//======================================  Inline methods.
inline bool
FileList::empty(void) const {
    return mFile.empty();
}

inline const char*
FileList::getFront(void) const {
    if (empty()) return 0;
    return mFile.front().c_str();
}

inline void 
FileList::pop_front(void) {
    mFile.pop_front();
}

inline unsigned long
FileList::size(void) const {
    return mFile.size();
}

#endif // FILELIST_HH
