//
//    Filter class FIR filter implementation.
//
#include "FIRFilter.hh"
#include "FIRdesign.hh"
#include <iostream>
#include <cstring>

//--------------------------------------  Filter design from poles, zeroes.
FIRFilter
design(double Hz, int nZeros, dComplex* Zeros) 
       throw(std::invalid_argument)
#ifdef RETCLAUSE
return r
#endif
{
    FIRFilter r;
    std::cerr << "Filter::design() has not been implemented. " << std::endl;
    std::cerr << "Please use dFirW() or dRemez() instead." << std::endl;
    return r;
}

//--------------------------------------  Filter design from poles, zeroes.
FIRFilter
dRemez(int N, double Hz, int nBand, const double* Bands, 
       const double* Func, const double* Weight) throw(std::invalid_argument)
#ifdef RETCLAUSE
return r
#endif
{
    if (Hz <= 0)  throw std::invalid_argument("Sampling Rate <= 0");
    if (nBand<=0) throw std::invalid_argument("nBand <= 0");

    FIRFilter r(N, Hz);
    double* fBand = new double [2*nBand];  // normalized bands

    for (int i=0 ; i<2*nBand ; i++) {
        fBand[i] = Bands[i] / Hz;
	if (fBand[i] < 0.0 || fBand[i] > 0.5) {
	    delete fBand;
	    throw std::invalid_argument("Band limits not in the range 0-Fny");
	}
    }

    int jtype = 1;
    double* pCoefs = new double[N+1];
    remez(N+1, jtype, nBand, 0, fBand, Func, Weight, pCoefs, 0, 0);
    r.setCoefs(N, pCoefs);
    delete[] pCoefs;
    delete [] fBand;
    return r;
}

#define N_WMODE 7
#define N_FMODE 4
const char* wmode[N_WMODE] = {"Rectangle", "Triangle", "Hamming", "GHamming", 
			      "Hanning", "Kaiser", "Cheby"};
const char* fmode[N_FMODE] = {"LowPass", "HighPass", "BandPass", "BandStop"};

FIRFilter
dFirW(int N, double Hz, const char* window, const char* type, double Flow, 
      double Fhigh, double Ripple, double dF) throw(std::runtime_error)
#ifdef RETCLAUSE
return r
#endif
{
    FIRFilter r;
    if (Hz <= 0) throw std::invalid_argument("Invalid sampling frequency.");

    //----------------------------------  Get the window mode
    int wm=-1;
    for (int i=0 ; i<N_WMODE ; i++) {
        if (!strcmp(wmode[i], window)) {
	    wm = i+1;
	    break;
	}
    }
    if (wm < 0) throw std::invalid_argument("Invalid window mode.");

    //----------------------------------  Get the filter mode
    int fm=-1;
    for (int i=0 ; i<N_FMODE ; i++) {
        if (!strcmp(fmode[i], type)) {
	    fm = i+1;
	    break;
	}
    }
    if (fm < 0) throw std::invalid_argument("Invalid filter mode.");

    //----------------------------------  Calculate the filter
    int Nf = N;
    double* h = new double[Nf];
    int rc = FirW(Nf, wm, fm, Flow/Hz, Fhigh/Hz, dF/Hz, Ripple, h);
    if (rc > 0) {  // if (rc==1) Nf was recalculated.
        delete[] h;
	h = new double[Nf];
	rc = FirW(Nf, wm, fm, Flow/Hz, Fhigh/Hz, dF/Hz, Ripple, h);
    }
    if (rc) throw std::runtime_error("Unable to construct filter");

    //----------------------------------  Fill in the filter
    int nset = (Nf+1)/2;
    for (int i=1 ; i<=nset ; i++) h[Nf-i] = h[nset-i];
    for (int i=1 ; i<=nset ; i++) h[i-1]  = h[Nf-i];
    r.setRate(Hz);
    r.setCoefs(Nf, h);
    delete[] h;
    return r;
}

//--------------------------------------  Estimate low-pass filter length
int
estLPLen(double Hz, double f1, double f2, double ripple, double atten) {
    double Fny = 0.5*Hz;
    if (f1<0 || f2<0 || f1>Fny || f2>Fny) {
        std::cerr << "Error: Band limits must be in the range 0.0-" 
	     << Fny << " Hz" << std::endl;
	return -1;
    }
    if (ripple<=0 || atten<=0) {
        std::cerr << "Filter ripple, attenuation (" << ripple << "," << atten
	     << ") must be >0" << std::endl;
	return -1;
    }
    double dr = log10(ripple);
    double da = log10(atten);
    double T  = -0.4278        -0.47610*da 
                -0.5941*dr     +0.07114*dr*da
                -0.00266*dr*dr +0.005309*dr*dr*da;
    double Kf = 11.01217 + 0.51244*(dr-da);
    double df = (f2-f1)/Hz;
    if (df<0.0) df = -df;
    double order = T/df - Kf*df;
    return int(ceil(order)) + 1;
}

//--------------------------------------  Estimate multi-band filter length.
int
estLen(double Hz, int nBand, double* Band, double* Rip) {
    int l, lmax=0;
    for (int j=1 ; j<nBand ; j++) {
	l = estLPLen(Hz, Band[2*j-1], Band[2*j], Rip[j-1], Rip[j]);
	if (l<=0) return -1;
	if (l>lmax) lmax = l;
	l = estLPLen(Hz, Band[2*j-1], Band[2*j], Rip[j], Rip[j-1]);
	if (l<=0) return -1;
	if (l>lmax) lmax = l;
    }
    return lmax;
}

